/*
 phencal.cpp

 Program to calculate phenological parameters from fitted spline models.

 This file is part of SPLITS, a framework for spline analysis of
 time series.

 Copyright (C) 2010-2015  Sebastian Mader

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <vector>
#include <gdal_priv.h>
#include <cpl_multiproc.h>
#define ARMA_NO_DEBUG
#include <armadillo>
#include "clargp.h"
#include "season.h"
#include "time_axis.h"
#include "dataset.h"
#include "spline.h"
#include "b_spline.h"
#include "common.h"
#include "phencal.h"

using namespace arma;
using namespace splits;

const char *in_file, *ot_file;
volatile int xSize, ySize, count, nseas, nband;
fvec xt;
int degree, num_knots, dim_basis;
float *knots;
bool peri = false;
Date d_beg, d_end;
std::vector<std::vector<Season> > seasons;
float period;
int indices = false;
bool raw = false;
bool southern = false;
volatile float fraction;
bool verbose = true;
int nthread = 1;
void *mutex;
static volatile int line = 0;
static volatile int pending = 1;

ClArg_t arglst[] =
{
  strlist_arg("C",1),
  real_arg("g",0),           //greenness parameter
  flag_arg("i",0),           //write indices from beginning instead of doy
  int_arg("M",0),            //number of threads in multiprocessing
  flag_arg("Qq",0),
  flag_arg("r",0),           //write raw time axis data
  flag_arg("S",0),           //southern hemisphere: shift skeleton by 172 days
  pos_arg("in_file"),
  pos_arg("ot_file"),
  last_arg
};

struct Buffer
{
  float *input;
  float *output;
  Spline *spline;
  std::vector<Season> seas;

  Buffer() : input(0), output(0), spline(0) {}

  Buffer(float* in, int off1, float* ot, int off2, Spline* s,
	 std::vector<Season>& v)
  {
    input = in+off1;
    output = ot+off2;
    spline = s;
    seas = v;
  }

  ~Buffer() {}
};

void greenness_start_end(Spline* spl,
			 float fraction,
			 const Season& seas,
			 float* gstart,
			 float* gend)
{
  int i;
  float y0s = seas.start.y + (seas.peak.y-seas.start.y)*fraction;
  float y0e = seas.end.y + (seas.peak.y-seas.end.y)*fraction;
  fvec xGs = spl->inv(y0s, seas.start.x, seas.end.x);
  float gs = seas.start.x;
  if (xGs.n_elem>0) {
    for (i=0; i<xGs.n_elem; i++)
      if (xGs(i)<seas.peak.x) gs = xGs(i);
  }
  fvec xGe = spl->inv(y0e, seas.start.x, seas.end.x);
  float ge = seas.end.x;
  if (xGe.n_elem>0) {
    for (i=xGe.n_elem-1; i>=0; i--)
      if (xGe(i)>seas.peak.x) ge = xGe(i);
  }
  *gstart = gs;
  *gend = ge;
}

void extract_flexpts(const Season& s,
		     const fvec& x,
		     const fvec& y,
		     const fvec& r,
		     int* p,
		     int* q)
{
  bool b = false;
  float rj;
  int i, j;
  for (i=0; i<x.n_elem; i++) {
    if (x[i]>=s.peak.x) break;
    if (x[i]>s.start.x) {
      if (b) {
	if (r[i]>=rj) {
	  j = i;
	  rj = r[i];
	}
      } else {
	j = i;
	rj = r[i];
	b = true;
      }
    }
  }
  if (b) *p=j; else *p=-1;

  b = false;
  for (i=0; i<x.n_elem; i++) {
    if (x[i]>=s.end.x) break;
    if (x[i]>s.peak.x) {
      if (b) {
	if (r[i]<=rj) {
	  j = i;
	  rj = r[i];
	}
      } else {
	j = i;
	rj = r[i];
	b = true;
      }
    }
  }
  if (b) *q=j; else *q=-1;
}

void extract_phenology(std::vector<Season>& seas,
		       Spline* spl,
		       float* buf,
		       float period,
		       float fraction,
		       const Date& d0)
{
  fvec flexX, flexY, flexR;
  bool flex = inflection_points(spl, flexX, flexY);
  if (flex) {
    flexR = spl->eval(flexX, 1);
  }

  int i, k, n=seas.size();
  for (i=0; i<n; i++)
  {
    Season s = seas[i];
    
    if (!s.peak) continue;

    if (fraction<(float)0) k = i*npar; else k = i*npar_g;

    int p, q;
    if (flex) extract_flexpts(s, flexX, flexY, flexR, &p, &q);

    buf[k++] = convert_index_to_doy(s.start.x, period, d0);
    if (p<0)
      buf[k++] = 0.0f;
    else
      buf[k++] = convert_index_to_doy(flexX[p], period, d0);
    buf[k++] = convert_index_to_doy(s.peak.x, period, d0);
    if (q<0)
      buf[k++] = 0.0f;
    else
      buf[k++] = convert_index_to_doy(flexX[q], period, d0);
    buf[k++] = convert_index_to_doy(s.end.x, period, d0);
    buf[k++] = s.start.y;
    if (p<0) buf[k++] = 0.0f; else buf[k++] = flexY[p];
    buf[k++] = s.peak.y;
    if (q<0) buf[k++] = 0.0f; else buf[k++] = flexY[q];
    buf[k++] = s.end.y;
    
    float duration = (s.end.x-s.start.x)*period;
    buf[k++] = duration;
    float latentVI = (s.start.y+s.end.y)/2.0;
    float amplitude = s.peak.y-latentVI;
    buf[k++] = amplitude;
    buf[k++] = latentVI;
    float latentInt = latentVI*duration;
    float totalInt = spl->integral(s.start.x,s.end.x)*period;
    float seasInt = totalInt-latentInt;
    buf[k++] = seasInt;
    buf[k++] = latentInt;
    buf[k++] = totalInt;
    if (p<0) buf[k++] = 0.0f; else buf[k++] = flexR[p]/period;
    if (q<0) buf[k++] = 0.0f; else buf[k++] = -flexR[q]/period;
    if (fraction>=(float)0) {
      float gs, ge;
      greenness_start_end(spl, fraction, s, &gs, &ge);
      float dy_gs = convert_index_to_doy(gs, period, d0);
      float dy_ge = convert_index_to_doy(ge, period, d0);
      float gs_val = spl->eval(gs);
      float ge_val = spl->eval(ge);
      buf[k++] = dy_gs;
      buf[k++] = dy_ge;
      buf[k++] = gs_val;
      buf[k++] = ge_val;
      buf[k++] = (ge-gs)*period;
      buf[k++] = (spl->integral(gs,ge)-(((gs_val+ge_val)/2.0)*(ge-gs)))*period;
      buf[k++] = (s.peak.y-gs_val)/((s.peak.x-gs)*period);
      buf[k++] = (s.peak.y-ge_val)/((ge-s.peak.x)*period);
    }
  }
}

void extract_phenology_index(std::vector<Season>& seas,
			     Spline* spl,
			     float* buf,
			     float period,
			     float fraction)
{
  fvec flexX, flexY, flexR;
  bool flex = inflection_points(spl, flexX, flexY);
  if (flex) {
    flexR = spl->eval(flexX, 1);
  }

  int i, k, n=seas.size();
  for (i=0; i<n; i++)
  {
    Season s = seas[i];

    if (!s.peak) continue;

    if (fraction<(float)0) k = i*npar; else k = i*npar_g;

    int p, q;
    if (flex) extract_flexpts(s, flexX, flexY, flexR, &p, &q);

    buf[k++] = s.start.x*period;
     if (p<0)
      buf[k++] = 0.0f;
    else
      buf[k++] = flexX[p]*period;
    buf[k++] = s.peak.x*period;
    if (q<0)
      buf[k++] = 0.0f;
    else
      buf[k++] = flexX[q]*period;
    buf[k++] = s.end.x*period;
    buf[k++] = s.start.y;
    if (p<0) buf[k++] = 0.0f; else buf[k++] = flexY[p];
    buf[k++] = s.peak.y;
    if (q<0) buf[k++] = 0.0f; else buf[k++] = flexY[q];
    buf[k++] = s.end.y;
    float duration = (s.end.x-s.start.x)*period;
    buf[k++] = duration;
    float latentVI = (s.start.y+s.end.y)/2.0;
    float amplitude = s.peak.y-latentVI;
    buf[k++] = amplitude;
    buf[k++] = latentVI;
    float latentInt = latentVI*duration;
    float totalInt = spl->integral(s.start.x,s.end.x)*period;
    float seasInt = totalInt-latentInt;
    buf[k++] = seasInt;
    buf[k++] = latentInt;
    buf[k++] = totalInt;
    if (p<0) buf[k++] = 0.0f; else buf[k++] = flexR[p]/period;
    if (q<0) buf[k++] = 0.0f; else buf[k++] = -flexR[q]/period;
    if (fraction>=(float)0) {
      float gs, ge;
      greenness_start_end(spl, fraction, s, &gs, &ge);
      float gs_val = spl->eval(gs);
      float ge_val = spl->eval(ge);
      buf[k++] = gs*period+1;
      buf[k++] = ge*period+1;
      buf[k++] = gs_val;
      buf[k++] = ge_val;
      buf[k++] = (ge-gs)*period;
      buf[k++] = (spl->integral(gs,ge)-(((gs_val+ge_val)/2.0)*(ge-gs)))*period;
      buf[k++] = (s.peak.y-gs_val)/((s.peak.x-gs)*period);
      buf[k++] = (s.peak.y-ge_val)/((ge-s.peak.x)*period);
    }
  }
}

void extract_phenology_raw(std::vector<Season>& seas,
			     Spline* spl,
			     float* buf,
			     float fraction)
{
  fvec flexX, flexY, flexR;
  bool flex = inflection_points(spl, flexX, flexY);
  if (flex) {
    flexR = spl->eval(flexX, 1);
  }

  int i, k, n=seas.size();
  for (i=0; i<n; i++)
  {
    Season s = seas[i];

    if (!s.peak) continue;

    if (fraction<(float)0) k = i*npar; else k = i*npar_g;

    int p, q;
    if (flex) extract_flexpts(s, flexX, flexY, flexR, &p, &q);

    buf[k++] = s.start.x;
     if (p<0)
      buf[k++] = 0.0f;
    else
      buf[k++] = flexX[p];
    buf[k++] = s.peak.x;
    if (q<0)
      buf[k++] = 0.0f;
    else
      buf[k++] = flexX[q];
    buf[k++] = s.end.x;
    buf[k++] = s.start.y;
    if (p<0) buf[k++] = 0.0f; else buf[k++] = flexY[p];
    buf[k++] = s.peak.y;
    if (q<0) buf[k++] = 0.0f; else buf[k++] = flexY[q];
    buf[k++] = s.end.y;
    float duration = s.end.x-s.start.x;
    buf[k++] = duration;
    float latentVI = (s.start.y+s.end.y)/2.0;
    float amplitude = s.peak.y-latentVI;
    buf[k++] = amplitude;
    buf[k++] = latentVI;
    float latentInt = latentVI*duration;
    float totalInt = spl->integral(s.start.x,s.end.x);
    float seasInt = totalInt-latentInt;
    buf[k++] = seasInt;
    buf[k++] = latentInt;
    buf[k++] = totalInt;
    if (p<0) buf[k++] = 0.0f; else buf[k++] = flexR[p];
    if (q<0) buf[k++] = 0.0f; else buf[k++] = -flexR[q];
    if (fraction>=(float)0) {
      float gs, ge;
      greenness_start_end(spl, fraction, s, &gs, &ge);
      float gs_val = spl->eval(gs);
      float ge_val = spl->eval(ge);
      buf[k++] = gs;
      buf[k++] = ge;
      buf[k++] = gs_val;
      buf[k++] = ge_val;
      buf[k++] = ge-gs;
      buf[k++] = spl->integral(gs,ge)-(((gs_val+ge_val)/2.0)*(ge-gs));
      buf[k++] = (s.peak.y-gs_val)/(s.peak.x-gs);
      buf[k++] = (s.peak.y-ge_val)/(ge-s.peak.x);
    }
  }
}

static void worker_func(void* p)
{
  Buffer *buf = (Buffer*)p;

  B_spline *bspl = (B_spline*)(buf->spline);

  std::vector<Season> seas = buf->seas;

  int i;
  CPLAcquireMutex(mutex, 1000.0);
  i = line++;
  CPLReleaseMutex(mutex);

  GDALDataset *in_dset;
  CPLAcquireMutex(mutex, 1000.0);
  in_dset = (GDALDataset*) GDALOpen(in_file, GA_ReadOnly);
  CPLReleaseMutex(mutex);

  in_dset->RasterIO(GF_Read,
		    0,i,xSize,1,
		    buf->input,xSize,1,
		    GDT_Float32,
		    count,
		    NULL,
		    count*sizeof(float),
		    count*xSize*sizeof(float),
		    sizeof(float));

  CPLAcquireMutex(mutex, 1000.0);
  GDALClose(in_dset);
  CPLReleaseMutex(mutex);

  int j, k;
  for (j=0; j<xSize; j++) {
    for (k=0; k<nband; k++) buf->output[j*nband+k] = 0.0f;
    fvec coefs = fvec((const float*)(buf->input+j*count), count);
    bool nonzero = false;
    for (k=0; k<coefs.n_elem; k++) {
      if (coefs[k]!=(float)0) nonzero = true;
      break;
    }
    if (nonzero) {
      bspl->coef(coefs);
      std::vector<Season>::iterator it;
      for (it=seas.begin(); it!=seas.end(); ++it) it->reset();
      identify_seasons(seas,bspl,1);
      if (indices) {
	extract_phenology_index(seas,bspl,buf->output+j*nband,period,fraction);
      } else if (raw) {
	extract_phenology_raw(seas,bspl,buf->output+j*nband,fraction);
      } else {
	extract_phenology(seas,bspl,buf->output+j*nband,period,fraction,d_beg);
      }
    }
  }

  GDALDataset *ot_dset;
  CPLAcquireMutex(mutex, 1000.0);
  ot_dset = (GDALDataset*) GDALOpen(ot_file, GA_Update);

  ot_dset->RasterIO(GF_Write,
		    0,i,xSize,1,
		    buf->output,xSize,1,
		    GDT_Float32,
		    nband,
		    NULL,
		    nband*sizeof(float),
		    nband*xSize*sizeof(float),
		    sizeof(float));

  GDALClose(ot_dset);
  CPLReleaseMutex(mutex);

  CPLAcquireMutex(mutex, 1000.0);
  pending--;
  CPLReleaseMutex(mutex);
}

int main(int argc, char* argv[])
{
  GDALAllRegister();
  CPLSetErrorHandler(CPLQuietErrorHandler);

  mutex = CPLCreateMutex();
  CPLReleaseMutex(mutex);

  parse_cl_args(argc, argv, arglst);

  int argsiz;
  void *parg;

  parg = get_cl_arg(arglst, "C", &argsiz);
  if (argsiz!=2) bail_out("Argument -C is invalid! Exit.");
  char **papch = (char**)parg;
  d_beg = Date(papch[0]);
  d_end = Date(papch[1]);

  parg = get_cl_arg(arglst, "g", &argsiz);
  if (parg) fraction = *(float*)parg; else fraction = (float)-1;

  parg = get_cl_arg(arglst, "i", &argsiz);
  if (argsiz>0) indices=true;

  parg = get_cl_arg(arglst, "r", &argsiz);
  if (argsiz>0) {
    if (indices) {
      bail_out("Only one of '-i' or '-r' may be given! Exit."); 
    } else {
      raw = true;
    }
  } 

  parg = get_cl_arg(arglst, "M", &argsiz);
  if (parg) {
    nthread = *(int*)parg;
    pending = nthread;
  }

  parg = get_cl_arg(arglst, "Qq", &argsiz);
  if (argsiz>0) verbose = false;

  parg = get_cl_arg(arglst, "S", &argsiz);
  if (argsiz>0) southern = true;

  in_file = (const char*)get_cl_arg(arglst, "in_file", &argsiz);
  ot_file = (const char*)get_cl_arg(arglst, "ot_file", &argsiz);

  Dataset d(in_file);

  knots = d.read_real_set("knot vector", &num_knots);
  if (knots==0) bail_out("Unable to get knot vector from header file! Exit.");

  if (!d.read_int("degree", &degree))
    bail_out("Unable to get degree from header file! Exit.");

  dim_basis = num_knots - (degree+1);

  int iper;
  d.read_int("periodic",&iper);
  if (iper==1) peri=true; else peri=false;

  xt = d.read_real_set("sites");

  period = (float)(d_end-d_beg);

  GDALDataset *in_dset = (GDALDataset*) GDALOpen(in_file,GA_ReadOnly);
  if (!in_dset) bail_out("Unable to open dataset for input! Exit.");
  xSize = in_dset->GetRasterXSize();
  ySize = in_dset->GetRasterYSize();
  count = in_dset->GetRasterCount();

  if (count!=dim_basis) {
    GDALClose(in_dset);
    bail_out("Invalid spline parameters in input dataset! Exit.");
  }

  int j;
  for (j=0; j<nthread; j++)
    seasons.push_back(create_skeleton(d_beg,
				      d_end,
				      southern,
				      0,
				      1));
  nseas = seasons[0].size();

  if (fraction<(float)0) nband = nseas*npar; else nband = nseas*npar_g;

  GDALDriver *drv = (GDALDriver*) GDALGetDriverByName("ENVI");

  char **opts = NULL;
  opts = CSLSetNameValue(opts, "INTERLEAVE", "BIL");
  opts = CSLSetNameValue(opts, "SUFFIX", "ADD");

  GDALDataset *ot_dset;
  ot_dset = drv->Create(ot_file,xSize,ySize,nband,GDT_Float32,opts);

  if (!ot_dset) {
    GDALClose(in_dset);
    bail_out("Unable to create output dataset! Exit.");
  }

  ot_dset->SetProjection(in_dset->GetProjectionRef());
  double adfGeoTransform[6];
  if (in_dset->GetGeoTransform(adfGeoTransform) == CE_None)
    ot_dset->SetGeoTransform(adfGeoTransform);

  GDALClose(ot_dset);
  GDALClose(in_dset);

  float *ibuf = new float[nthread*xSize*count];
  float *obuf = new float[nthread*xSize*nband];

  int offs_in = xSize*count;
  int offs_ot = xSize*nband;

  Spline **spl = new Spline*[nthread];

  for (j=0; j<nthread; j++) {
    spl[j] = new B_spline(knots,dim_basis,degree+1,peri);
  }

  Buffer *buffers = new Buffer[nthread];
  for (j=0; j<nthread; j++) {
    buffers[j] = Buffer(ibuf, j*offs_in, obuf, j*offs_ot, spl[j], seasons[j]);
  }

  int i, n;
  for (i=0; i<ySize; i+=nthread) {
    if ((i+nthread)>=ySize)
      n = ySize-i;
    else
      n = nthread;
    pending = n;
    for (j=0; j<n; j++) {
      CPLCreateThread(worker_func, (void*)&buffers[j]);
    }
    while (pending > 0) CPLSleep(0.5);
    if (verbose) report_progress(nthread, ySize);
    CPLReleaseMutex(mutex);
  }

  int np;
  if (fraction<0.0f) np=npar; else np=npar_g; 

  d = Dataset(ot_file);
  d.set_band_names(seasons[0], np, param_names);

  d.mk_meta_files(np, param_names, nseas);
  d.mk_meta_files(seasons[0], np);

  delete [] buffers;

  delete [] ibuf;
  delete [] obuf;

  delete [] spl;

  return 0;
}
