/*
 season.cpp

 Analyze seasons and phenology using a spline model.

 The SPLITS splview program to fit and visualize spline models to
 remotely sensed time series images.

 This file is part of SPLITS, a framework for spline analysis of
 time series.

 Copyright (C) 2010-2015  Sebastian Mader

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "season.h"

using namespace std;
using namespace arma;
using namespace splits;

void Season::reset()
{
  start.x = 0;
  start.y = 0;
  peak.x = 0;
  peak.y = 0;
  end.x = 0;
  end.y = 0;
}


void splits::extreme_points(Spline* spl,
			    fvec& xmn,
			    fvec& ymn,
			    fvec& xmx,
			    fvec& ymx)
{
  fvec x0 = spl->roots(1);
  fvec y0 = spl->eval(x0);
  fvec ydblpr0 = spl->eval(x0,2);
  int i, nmn=0, nmx=0;
  for (i=0; i<ydblpr0.n_elem; i++) {
    if (ydblpr0(i)<(float)0) nmx++;
    else if (ydblpr0(i)>(float)0) nmn++;
  }
  xmn = fvec(nmn); ymn = fvec(nmn);
  xmx = fvec(nmx); ymx = fvec(nmx);
  int imn=0, imx=0;
  for (i=0; i<ydblpr0.n_elem; i++) {
    if (ydblpr0(i)<(float)0) {
      xmx(imx) = x0(i);
      ymx(imx) = y0(i);
      imx++;
    } else if (ydblpr0(i)>(float)0) {
      xmn(imn) = x0(i);
      ymn(imn) = y0(i);
      imn++;
    }
  }
}


vector<int> splits::extreme_points(Spline* spl, fvec& x, fvec& y)
{
  fvec x0 = spl->roots(1);
  fvec y0 = spl->eval(x0);
  fvec ydblpr0 = spl->eval(x0,2);

  int i, n=0;
  for (i=0; i<ydblpr0.n_elem; i++) {
    if (ydblpr0[i]<(float)0) n++;
    else if (ydblpr0[i]>(float)0) n++;
  }

  x = fvec(n);
  y = fvec(n);
  vector<int> sgn(n);

  int j=0;
  for (i=0; i<ydblpr0.n_elem; i++) {
    if (ydblpr0[i]<(float)0) {
      x[j] = x0[i];
      y[j] = y0[i];
      sgn[j++]=-1;
    } else if (ydblpr0[i]>(float)0) {
      x[j] = x0[i];
      y[j] = y0[i];
      sgn[j++]=1;
    }
  }

  return sgn;
}


bool splits::inflection_points(Spline* spl, arma::fvec& x, arma::fvec& y)
{
  if (spl->degree()<3) {
    x = fvec();
    y = fvec();
    return false;
  }

  fvec x0 = spl->roots(2);
  fvec y0 = spl->eval(x0);
  fvec ytrplpr0 = spl->eval(x0,3);

  int i, n=0;
  for (i=0; i<ytrplpr0.n_elem; i++) {
    if (ytrplpr0[i]!=(float)0) n++;
  }

  x = fvec(n);
  y = fvec(n);

  int j=0;
  for (i=0; i<ytrplpr0.n_elem; i++) {
    if (ytrplpr0[i]!=(float)0) {
      x[j] = x0[i];
      y[j] = y0[i];
      j++;
    }
  }

  return true;
}


static void seasons1(fvec& xmn, fvec& ymn, vector<Season>& seasv, float imin,
		     float imax)
{
  int i, j;

  /* Find the smallest minimum in between two consecutive peaks. Define
     it to be the end of the former and the start of the latter season */
  int n1=seasv.size()-1;
  float left, right, mn_x, mn_y;
  bool flag;
  for (i=0; i<n1; i++) {
    left = seasv[i].peak.x;
    right = seasv[i+1].peak.x;
    flag = false;
    for (j=0; j<xmn.n_elem; j++) {
      if ((xmn(j)>=left) && (xmn(j)<=right)) {
	if (!flag) {
	  mn_x = xmn(j);
	  mn_y = ymn(j);
	  flag = true;
	} else {
	  if (ymn(j)<mn_y) {
	    mn_x = xmn(j);
	    mn_y = ymn(j);
	  }
	}
      }
    }
    if (flag) {
      DataPt p(mn_x,mn_y);
      seasv[i].end = p;
      seasv[i+1].start = p;
    }
  }

  /* Find the end of the last season of the time series. The end is
   located between the peak of the last season and imax or the smallest
   minimum following imax */
  left = seasv[n1].peak.x;
  right = imax;
  flag = false;
  for (j=xmn.n_elem-1; j>=0; j--) {
    if (xmn(j)<=imax) {
      break;
    } else {
      if (!flag) {
	right = xmn(j);
	mn_y = ymn(j);
	flag = true;
      } else {
	right = xmn(j);
	mn_y = ymn(j);
      }
    }
  }
  flag = false;
  for (j=0; j<xmn.n_elem; j++) {
    if ((xmn(j)>=left) && (xmn(j)<=right)) {
      if (!flag) {
	mn_x = xmn(j);
	mn_y = ymn(j);
	flag = true;
      } else {
	if (ymn(j)<mn_y) {
	  mn_x = xmn(j);
	  mn_y = ymn(j);
	}
      }
    }
  }
  if (flag) {
    DataPt p(mn_x,mn_y);
    seasv[n1].end = p;
  }

  /* Find the start of the first season of the time series. The start is
   located between imin or the smallest minimum before imin and the peak
   of the first season */
  right = seasv[0].peak.x;
  left = imin;
  flag = false;
  for (j=0; j<xmn.n_elem; j++) {
    if (xmn(j)>=imin) {
      break;
    } else {
      if (!flag) {
	left = xmn(j);
	mn_y = ymn(j);
	flag = true;
      } else {
	left = xmn(j);
	mn_y = ymn(j);
      }
    }
  }
  flag = false;
  for (j=0; j<xmn.n_elem; j++) {
    if ((xmn(j)>=left) && (xmn(j)<=right)) {
      if (!flag) {
	mn_x = xmn(j);
	mn_y = ymn(j);
	flag = true;
      } else {
	if (ymn(j)<mn_y) {
	  mn_x = xmn(j);
	  mn_y = ymn(j);
	}
      }
    }
  }
  if (flag) {
    DataPt p(mn_x,mn_y);
    seasv[0].start = p;
  }
}

void splits::identify_seasons(vector<Season>& seas,
			      Spline* spl,
			      float imax,
			      float range,
			      bool bigT,
			      float low,
			      float high,
			      bool atBeg,
			      bool atEnd)
{
  fvec xmn, ymn, xmx, ymx;
  extreme_points(spl, xmn, ymn, xmx, ymx);

  // Initialize every season and find the biggest peak
  int j;
  bool flag;
  float peak_x, peak_y;
  vector<Season>::iterator it;
  for (it=seas.begin(); it!=seas.end(); ++it) {
    it->reset();
    flag = false;
    for (j=0; j<xmx.n_elem; j++) {
      float mx = xmx[j];
      if (mx>=it->startIndex && mx<=it->endIndex) {
	if (flag) {
	  if (ymx[j]>peak_y) {
	    peak_x = mx;
	    peak_y = ymx[j];
	  }
	} else {
	  peak_x = mx;
	  peak_y = ymx[j];
	  flag = true;
	}
      }
    }
    if (flag) {
      it->peak.x = peak_x;
      it->peak.y = peak_y;
    }
  }

  seasons1(xmn, ymn, seas, seas.begin()->startIndex, (seas.end()-1)->endIndex);

  // Reset (zero out) incomplete seasons
  for (it=seas.begin(); it!=seas.end(); ++it) {
    if (!(it->end)) it->reset();
  }
}
